@@ -465,7 +465,7 @@ class WindowDensityModel(BaseModel):
 
     def _call_scoring(self, df=None, target_metric=None, anomaly_scores_gamma_alpha=None, anomaly_scores_gamma_loc=None,
                       anomaly_scores_gamma_beta=None, baseline=None, detrend_order=None, detrend_method=None,
-                      agg_data_model=None, detection_method=None, attributes=None):
+                      agg_data_model=None, detection_method=None, attributes=None, agg_data=None):
         """
         This function generates the anomaly flag and and probability for the scoring window.
         :param pandas.DataFrame df: Input training data frame.
@@ -479,6 +479,7 @@ class WindowDensityModel(BaseModel):
         :param luminaire.model.lad_structural.LADStructuralModel agg_data_model: Prediction model for aggregated data.
         :param str detection_method: Selects between "kldiv" or "sign_test" distance method.
         :param attributes: Model attributes.
+        :param agg_data: Aggregated Data per day.
         :return: Returns the anomaly flag with the corresponding anomaly probability.
         :rtype: tuple(bool, float, dict)
         """
@@ -492,17 +493,21 @@ class WindowDensityModel(BaseModel):
                                                                        detrend_order=detrend_order,
                                                                        detrend_method=detrend_method,
                                                                        agg_data_model=agg_data_model,
-                                                                       detection_method=detection_method)
+                                                                       detection_method=detection_method,
+                                                                       agg_data=agg_data)
 
         return is_anomaly, prob_of_anomaly, attributes
 
     def _get_result(self, input_df=None, detrend_order=None, agg_data_model=None, value_column=None,
                     detrend_method=None, baseline_type=None, detection_method=None, baseline=None,
-                    anomaly_scores_gamma_alpha=None, anomaly_scores_gamma_loc=None, anomaly_scores_gamma_beta=None):
+                    anomaly_scores_gamma_alpha=None, anomaly_scores_gamma_loc=None, anomaly_scores_gamma_beta=None,
+                    agg_data=None):
         """
         The function scores the scoring window for anomalies based on the training metrics and the baseline
         :param pandas.DataFrame input_df: Input data containing the training and the scoring data.
-        :param int detrend_order: The order of detrending based on MA or differencing method.
+        :param int detrend_order: The non-negative order of detrending based on Modeling or differencing method. When
+        the detrend_order > 0, corresponding detrending need to be performed using the method specified in the model
+        config.
         :param luminaire.model.lad_structural.LADStructuralModel agg_data_model: Prediction model for aggregated data.
         :param str value_column: Column containing the values.
         :param str detrend_method: Selects between "modeling" or "diff" detrend method.
@@ -512,6 +517,7 @@ class WindowDensityModel(BaseModel):
         :param float anomaly_scores_gamma_alpha: Gamma fit alpha parameter.
         :param float anomaly_scores_gamma_loc: Gamma fit location parameter.
         :param float anomaly_scores_gamma_beta: Gamma fit beta parameter.
+        :param agg_data: Aggregated Data per day.
         :return: Returns the anomaly flag with the corresponding anomaly probability.
         :rtype: tuple(bool, float)
         """
@@ -528,6 +534,10 @@ class WindowDensityModel(BaseModel):
 
         is_anomaly = False
         execution_data = input_df[value_column]
+        adjusted_execution_data = []
+        prob_of_anomaly = []
+        len_req_agg_data_model = 42     # Setting a hard threshold to have predictions from aggregated data
+        # for stationarity adjustment
 
         if detrend_method == 'diff':
             # Obtain the execution data and perform the necessary differencing
@@ -539,23 +549,51 @@ class WindowDensityModel(BaseModel):
             dates_freq_dist = dict(collections.Counter(idx))
             scoring_datetime = str(max(dates_freq_dist.items(), key=operator.itemgetter(1))[0])
             execution_data_avg = np.mean(execution_data)
-            try:
-                data_adjust_forecast = agg_data_model.score(execution_data_avg, scoring_datetime)['Prediction'] \
-                    if agg_data_model else 1.0
-            except:
-                data_adjust_forecast = 1
-            adjusted_execution_data = execution_data / data_adjust_forecast
+            # If detrending is needed, we scale the scoring data accordingly using the agg_dat_model forecast
             if detrend_order > 0:
-                adjusted_execution_data = adjusted_execution_data.to_list()
+                snapshot_len_max = min(len(agg_data), len_req_agg_data_model)
+                agg_data_trunc = np.array(agg_data)[:, 1][-snapshot_len_max:]
+                data_adjust_forecast = []
+                try:
+                    # Setting the data adjustment window of the original data using the predictions and the CILower and
+                    # CIUpper keeping the prediction uncertainty of the agg_model in mind
+                    if agg_data_model and len(agg_data) > len_req_agg_data_model:
+                        score = agg_data_model.score(execution_data_avg, scoring_datetime)
+                        data_adjust_forecast.append(score['Prediction'])
+                        data_adjust_forecast.append(score['CILower'])
+                        data_adjust_forecast.append(score['CIUpper'])
+                    else:
+                        data_adjust_forecast.append(np.median(agg_data_trunc))
+                        data_adjust_forecast.append(np.percentile(agg_data_trunc, 5))       # setting a 2-sigma limit
+                        data_adjust_forecast.append(np.percentile(agg_data_trunc, 95))       # setting a 2-sigma limit
+                except:
+                    # If the scoring for the agg_data_model fails for some reason, we use the latest agg_data for the
+                    # detrending adjustment
+                    data_adjust_forecast.append(np.median(agg_data_trunc))
+                    data_adjust_forecast.append(np.percentile(agg_data_trunc, 5))       # setting a 2-sigma limit
+                    data_adjust_forecast.append(np.percentile(agg_data_trunc, 95))       # setting a 2-sigma limit
+                for i in range(3):
+                    if data_adjust_forecast[i] != 0:
+                        adjusted_execution_data.append((execution_data / data_adjust_forecast[i]).tolist())
             else:
-                adjusted_execution_data = list(adjusted_execution_data)
+                adjusted_execution_data = list(execution_data)
 
         # Kl divergence based anomaly detection
         if detection_method == "kldiv":
-            current_anomaly_score = self._distance_function(data=adjusted_execution_data,
-                                                            called_for="scoring", baseline=baseline)
-            prob_of_anomaly = st.gamma.cdf(current_anomaly_score, anomaly_scores_gamma_alpha,
-                                           anomaly_scores_gamma_loc, anomaly_scores_gamma_beta)
+            if detrend_order > 0:
+                prob_of_anomaly = []
+                for i in range(3):
+                    current_anomaly_score = self._distance_function(data=adjusted_execution_data[i],
+                                                                    called_for="scoring", baseline=baseline)
+                    prob_of_anomaly.append(st.gamma.cdf(current_anomaly_score, anomaly_scores_gamma_alpha,
+                                                        anomaly_scores_gamma_loc, anomaly_scores_gamma_beta))
+                prob_of_anomaly = np.min(prob_of_anomaly)
+            else:
+                current_anomaly_score = self._distance_function(data=adjusted_execution_data,
+                                                                called_for="scoring", baseline=baseline)
+                prob_of_anomaly = st.gamma.cdf(current_anomaly_score, anomaly_scores_gamma_alpha,
+                                               anomaly_scores_gamma_loc, anomaly_scores_gamma_beta)
+
             if 1 - prob_of_anomaly < self.sig_level:
                 is_anomaly = True
         # Sign test based anomaly detection
@@ -572,20 +610,25 @@ class WindowDensityModel(BaseModel):
             # based test for the past standard deviations to detect anomalies
             elif baseline_type == "aggregated":
                 baseline_sds = np.array(baseline).std(1).tolist()
-                baseline_execution_data = copy.copy(baseline)
-                baseline_execution_data.append(adjusted_execution_data)
-                pca = PCA()
-                scores = pca.fit_transform(StandardScaler().fit_transform(baseline_execution_data))
-                robust_cov = MinCovDet().fit(scores[:, :3])
-                mahalanobis_distance = robust_cov.mahalanobis(scores[:, :3])
-                pvalue_mahalanobis = 1 - st.chi2.cdf(mahalanobis_distance[-1],
-                                                     np.array(baseline_execution_data).shape[1])
+                if detrend_order == 0:
+                    # crearing a 2d list to make it easy to loop through in the following for loop
+                    adjusted_execution_data = [adjusted_execution_data]
+                for current_adjusted_data in adjusted_execution_data:
+                    baseline_execution_data = copy.copy(baseline)
+                    baseline_execution_data.append(current_adjusted_data)
+                    pca = PCA()
+                    scores = pca.fit_transform(StandardScaler().fit_transform(baseline_execution_data))
+                    robust_cov = MinCovDet().fit(scores[:, :3])
+                    mahalanobis_distance = robust_cov.mahalanobis(scores[:, :3])        # getting the top 3 dimensions
+                    pvalue_mahalanobis = 1 - st.chi2.cdf(mahalanobis_distance[-1],
+                                                         np.array(baseline_execution_data).shape[1])
 
-                gamma_alpha, gamma_loc, gamma_beta = st.gamma.fit(baseline_sds)
-                pvalue_gamma = 1 - st.gamma.cdf(np.std(adjusted_execution_data), gamma_alpha, gamma_loc, gamma_beta)
-                if pvalue_mahalanobis < self.sig_level or pvalue_gamma < self.sig_level:
-                    is_anomaly = True
-                prob_of_anomaly = 1 - min(pvalue_mahalanobis, pvalue_gamma)
+                    gamma_alpha, gamma_loc, gamma_beta = st.gamma.fit(baseline_sds)
+                    pvalue_gamma = 1 - st.gamma.cdf(np.std(current_adjusted_data), gamma_alpha, gamma_loc, gamma_beta)
+                    if pvalue_mahalanobis < self.sig_level or pvalue_gamma < self.sig_level:
+                        is_anomaly = True
+                    prob_of_anomaly.append(1 - min(pvalue_mahalanobis, pvalue_gamma))
+                prob_of_anomaly = np.min(prob_of_anomaly)
 
         return is_anomaly, prob_of_anomaly
 
@@ -662,6 +705,7 @@ class WindowDensityModel(BaseModel):
         baseline = self._params['Baseline'][opt_timestamp]
         detrend_order = self._params['NonStationarityOrder'][opt_timestamp]
         agg_data_model = self._params['AggregatedDataModel'][opt_timestamp]
+        agg_data = self._params['AggregatedData'][opt_timestamp]
 
         is_anomaly, prob_of_anomaly, attributes = self._call_scoring(df=data,
                                                                      target_metric=target_metric,
@@ -672,7 +716,8 @@ class WindowDensityModel(BaseModel):
                                                                      detrend_order=detrend_order,
                                                                      detrend_method=detrend_method,
                                                                      agg_data_model=agg_data_model,
-                                                                     detection_method=detection_method)
+                                                                     detection_method=detection_method,
+                                                                     agg_data=agg_data)
 
         result = {'Success': True,
                   'ConfLevel': float(1.0 - self.sig_level) * 100,
@@ -687,7 +732,8 @@ class WindowDensityModel(BaseModel):
                                     value_column=None, called_for=None,
                                     anomaly_scores_gamma_alpha=None, anomaly_scores_gamma_loc=None,
                                     anomaly_scores_gamma_beta=None, detrend_order=None, baseline=None,
-                                    detrend_method=None, agg_data_model=None, past_model=None, detection_method=None):
+                                    detrend_method=None, agg_data_model=None, past_model=None, detection_method=None,
+                                    agg_data=None):
         """
         This function detects anomaly given a training and a scoring window.
 
@@ -704,6 +750,7 @@ class WindowDensityModel(BaseModel):
         :param luminaire.model.lad_structural.LADStructuralModel agg_data_model: Prediction model for aggregated data.
         :param luminaire.model.window_density.WindowDensityModel past_model: Past stored window density model.
         :param str detection_method: Selects between "kldiv" or "sign_test" distance method.
+        :param agg_data: Aggregated Data per day.
         :return: Anomaly flag with the corresponding probability of anomaly.
         :rtype: tuple(bool, float)
 
@@ -736,4 +783,5 @@ class WindowDensityModel(BaseModel):
                                     baseline=baseline,
                                     anomaly_scores_gamma_alpha=anomaly_scores_gamma_alpha,
                                     anomaly_scores_gamma_loc=anomaly_scores_gamma_loc,
-                                    anomaly_scores_gamma_beta=anomaly_scores_gamma_beta)
+                                    anomaly_scores_gamma_beta=anomaly_scores_gamma_beta,
+                                    agg_data=agg_data)
